import java.util.*;
import negotiator.*;
import negotiator.actions.*;
import negotiator.issue.*;

/**
 * @author W.Pasman Some improvements over the standard SimpleAgent.
 */
public class AgentMR extends Agent {
	private Action actionOfPartner = null;
	private ArrayList<Bid> bidRunk = new ArrayList<Bid>();
	private ArrayList<Double> observationUtility = new ArrayList<Double>();
	private HashMap<Bid, Double> bidTables = new HashMap<Bid, Double>();
	private static final double MINIMUM_ACCEPT_P = 0.965;
	private static boolean firstOffer = true;
	private static boolean forecastTime = true;
	private static boolean discountFactor;
	private static double minimumBidUtility;
	private static double minimumOffereDutil;
	private static Bid previousPartnerBid = null;
	private static Bid offereMaxBid = null;
	private static double offereMaxUtility;
	private static double firstOffereUtility;
	private int currentBidNumber = 0;
	private int lastBidNumber = 1;
	private double sigmoidGain;
	private double sigmoidX;
	private double reservation = 0.0;
	private double alpha;
	private double percent;
	private double p = 0.90;

	/**
	 * init is called when a next session starts with the same opponent.
	 */
	public void init() {
		try {
			getDiscountFactor();
			getReservationFactor();
			updateMinimumBidUtility(0);
			Bid b = utilitySpace.getMaxUtilityBid();
			bidTables.put(b, getUtility(b));
			bidRunk.add(b);
			if (discountFactor) { // 割引率がある場合
				sigmoidGain = -3;
				percent = 0.55;
			} else { // 割引率がない場合
				sigmoidGain = -5;
				percent = 0.70;
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	public static String getVersion() {
		return "1.2";
	}

	public void ReceiveMessage(Action opponentAction) {
		actionOfPartner = opponentAction;
	}

	public Action chooseAction() {
		Action action = null;

		try {
			if (actionOfPartner == null) {
				// 初回時は最高値をBid
				action = new Offer(getAgentID(),
						utilitySpace.getMaxUtilityBid());
			}
			if (actionOfPartner instanceof Offer) {
				Bid partnerBid = ((Offer) actionOfPartner).getBid();

				// get current time
				double time = timeline.getTime();

				double offeredutil; // 相手のBidUtility(常に割引率の影響を受けていないものを取得)
				if (discountFactor) {
					offeredutil = getUtility(partnerBid)
							* (1 / Math.pow(utilitySpace.getDiscountFactor(),
									time));
				} else {
					offeredutil = getUtility(partnerBid);
				}

				// 初回オファー処理
				if (firstOffer) {
					previousPartnerBid = partnerBid; // 相手の過去Bidをセット
					offereMaxBid = partnerBid; // 相手の最大Bidをセット
					offereMaxUtility = offeredutil; // 相手の最大BidUtilをセット
					firstOffereUtility = offeredutil; // 相手の初期BidUtilをセット
					observationUtility.add(offeredutil); // addObservation
					if (offeredutil > 0.5) { // 逓減率セット
						p = 0.90;
					} else {
						p = 0.80;
					}
					firstOffer = !firstOffer; // 初回オファー処理終了
				}

				updateMinimumBidUtility(time); // 閾値(MBU)処理

				// Bid移動処理
				if (partnerBid.equals(previousPartnerBid)) { // 相手の提示が変わらない場合
					if (currentBidNumber > 0 && Math.random() > 0.65) {
						currentBidNumber--; // 確率的にBidを移動
					}
				}

				// 相手の最大Bidを更新
				if (offeredutil > offereMaxUtility) {
					offereMaxBid = partnerBid;
					offereMaxUtility = offeredutil;
					// addObservation
					observationUtility.add(offeredutil);
					if ((time > 0.5) && !discountFactor) {
						newupdateSigmoidFunction();
					}
				}

				// forecasting処理
				if ((time > 0.5) && forecastTime) {
					updateSigmoidFunction();
					forecastTime = !forecastTime;
				}

				double P = Paccept(offeredutil, time);

				// Accept処理
				// 1. AcceptPの下限を超えた場合
				// 2. 閾値(MBU)を超えたBidが提示された場合
				// 3. 既にこちらがBidをしたBidが提示された場合
				if ((P > MINIMUM_ACCEPT_P) || (offeredutil > minimumBidUtility)
						|| bidRunk.contains(partnerBid)) {
					action = new Accept(getAgentID());
				} else {
					// 相手提示の最大BidUtiltyがMBU以上ならそれをBidし続ける
					// 早期合意形成を図る
					if (offereMaxUtility > minimumBidUtility) {
						action = new Offer(getAgentID(), offereMaxBid);
					}
					// 最終Bid処理
					// 留保価格がなければ合意形成を最優先とする
					else if (time > 0.985) {
						// 相手の最大BidをBidし、合意形成
						if (offereMaxUtility > reservation) {
							action = new Offer(getAgentID(), offereMaxBid);
						} else { // 相手の最大BidUtilが留保価格以下なら自分の最大の譲歩Bidを行う
							action = new Offer(getAgentID(),
									bidRunk.get(bidRunk.size() - lastBidNumber));
							lastBidNumber++;
						}
					}
					// 通常Bid処理
					// 相手の提示Bidが一定以上を超えた場合
					else if (offeredutil > minimumOffereDutil) {
						HashMap<Bid, Double> getBids = getBidTable(1);
						// 相手のBidの近くにMBU以上のBidが存在した場合
						if (getBids.size() >= 1) {
							// BidTableを初期化・再構築
							currentBidNumber = 0;
							bidRunk.clear();
							bidTables = getBids;
							sortBid(getBids); // Sort BidTable
						} else { // 通常のBid検索
							getBids = getBidTable(2);
							if (getBids.size() >= 1) {
								sortBid(getBids); // Sort BidTable
								Bid maxBid = getMaxBidUtility(getBids);
								// 見つけたBidのうち最大の番号を割り当て
								currentBidNumber = bidRunk.indexOf(maxBid);
							}
						}
						action = new Offer(getAgentID(),
								bidRunk.get(currentBidNumber));
						// 次のBidへと移動
						if (currentBidNumber + 1 < bidRunk.size()) {
							currentBidNumber++;
						}
					} else {
						HashMap<Bid, Double> getBids = getBidTable(2);
						if (getBids.size() >= 1) {
							sortBid(getBids); // Sort BidTable
							Bid maxBid = getMaxBidUtility(getBids);
							currentBidNumber = bidRunk.indexOf(maxBid);
						}
						action = new Offer(getAgentID(),
								bidRunk.get(currentBidNumber));
						if (currentBidNumber + 1 < bidRunk.size()) {
							currentBidNumber++;
						} else {
							currentBidNumber = 0;
						}
					}
				}
				// 相手の一つ前のBidを記憶しておく
				previousPartnerBid = partnerBid;
			}
		} catch (Exception e) {
			e.printStackTrace();
			action = new Accept(getAgentID()); // best guess if things go wrong.
		}
		return action;
	}

	// 留保価格
	private void getReservationFactor() {
		if (utilitySpace.getReservationValue() != null) {
			reservation = utilitySpace.getReservationValue();
		}
	}

	// 割引率の有無
	private void getDiscountFactor() {
		if (utilitySpace.getDiscountFactor() > 0.0) {
			discountFactor = true; // 割引率有り
		} else
			discountFactor = false; // 割引率無し
	}

	// forecasting処理
	private void newupdateSigmoidFunction() {
		double latestObservation = observationUtility.get(observationUtility.size() - 1);
		double concessionPercent = Math.abs(latestObservation - firstOffereUtility) / (1.0 - firstOffereUtility);
		double modPercent = Math.abs(minimumOffereDutil - firstOffereUtility) / (1.0 - firstOffereUtility);

		if (modPercent < concessionPercent) {
			percent = concessionPercent;
		}
	}

	private void updateSigmoidFunction() {
		int observationSize = observationUtility.size();
		double latestObservation = observationUtility.get(observationSize - 1); // 最新の相手BidUtil
		double concessionPercent = Math.abs(latestObservation - firstOffereUtility) / (1.0 - firstOffereUtility);

		if (discountFactor) {
			// Agent譲歩なしと判断
			if ((concessionPercent < 0.20) ||
					(observationSize < 3)) {
				percent = 0.35;
				sigmoidGain = -2;
			} else {
				percent = 0.45;
			}
		} else {
			// Agent譲歩なしと判断
			if ((concessionPercent < 0.20) ||
					(observationSize < 3)) {
				percent = 0.50;
				sigmoidGain = -4;
			} else if (concessionPercent > 0.60) { // Agent譲歩大
				percent = 0.80;
				sigmoidGain = -6;
			} else {
				percent = 0.60;
			}
		}
	}

	// 相手の提示Bidから最大のものを返す
	private Bid getMaxBidUtility(HashMap<Bid, Double> bidTable) {
		Double maxBidUtility = 0.0;
		Bid maxBid = null;
		for (Bid b : bidTable.keySet()) {
			if (getUtility(b) > maxBidUtility) {
				maxBidUtility = getUtility(b);
				maxBid = b;
			}
		}
		return maxBid;
	}

	private void updateMinimumBidUtility(double time) {

		alpha = (1.0 - firstOffereUtility) * percent; // 初期Utilityとの差x%

		// 相手の初期Utility+alphaをMBU下限としたsigmoidFuncの定義域を計算
		double mbuInfimum = firstOffereUtility + alpha;
		if (mbuInfimum >= 1.0) {
			mbuInfimum = 0.999; // 0除算対策
		} else if (mbuInfimum <= 0.70) {
			mbuInfimum = 0.70; // MOD下限
		}
		sigmoidX = 1 - ((1 / sigmoidGain) * Math.log(mbuInfimum / (1 - mbuInfimum)));

		minimumBidUtility = 1 - (1 / (1 + Math.exp(sigmoidGain
				* (time - sigmoidX)))); // シグモイド関数で推移

		if (minimumBidUtility < reservation) { // 留保価格以下には下げない
			minimumBidUtility = reservation;
		}

		// 相手許容Bid下限(MOD)
		minimumOffereDutil =  minimumBidUtility * p;
	}

	/**
	 * BidTableを降順ソート
	 *
	 * @param bidTable
	 */
	private void sortBid(final HashMap<Bid, Double> getBids) {

		for (Bid bid : getBids.keySet()) {
			bidTables.put(bid, getUtility(bid));
			bidRunk.add(bid); // Add bidRunk
		}

		// Bidソート処理
		Collections.sort(bidRunk, new Comparator<Bid>() {
			@Override
			public int compare(Bid o1, Bid o2) {
				return (int) Math.ceil(-(bidTables.get(o1) - bidTables.get(o2)));
			}
		});
	}

	private Bid clone(Bid source) throws Exception {
		HashMap<Integer, Value> hash = new HashMap<Integer, Value>();
		for (Issue i : utilitySpace.getDomain().getIssues()) {
			hash.put(i.getNumber(), source.getValue(i.getNumber()));
		}
		return new Bid(utilitySpace.getDomain(), hash);
	}

	/**
	 * Bidから近いかつUtilityが高いものを探索
	 *
	 * @param maxBid
	 * @return
	 * @throws Exception
	 */
	private HashMap<Bid, Double> getBidTable(int flag) throws Exception {
		HashMap<Bid, Double> getBids = new HashMap<Bid, Double>();
		Random randomnr = new Random();
		ArrayList<Issue> issues = utilitySpace.getDomain().getIssues();
		Bid standardBid = null;

		for (Issue lIssue : issues) {
			switch (lIssue.getType()) {
			case DISCRETE:
				IssueDiscrete lIssueDiscrete = (IssueDiscrete) lIssue;
				for (ValueDiscrete value : lIssueDiscrete.getValues()) {
					if (flag == 0) {
						standardBid = utilitySpace.getMaxUtilityBid(); // 自分の最高値
					} else if (flag == 1) {
						standardBid = ((Offer) actionOfPartner).getBid(); // 相手のBid
					} else {
						standardBid = bidRunk.get(currentBidNumber);
					}
					standardBid = clone(standardBid);
					standardBid.setValue(lIssue.getNumber(), value);
					double utility = getUtility(standardBid);
					if ((utility > minimumBidUtility)
							&& (!bidRunk.contains(standardBid))) {
						getBids.put(standardBid, utility);
					}
				}
				break;
			case REAL:
				IssueReal lIssueReal = (IssueReal)lIssue;
				int optionInd = randomnr.nextInt(lIssueReal.getNumberOfDiscretizationSteps()-1);
				Value pValue = new ValueReal(lIssueReal.getLowerBound() + (lIssueReal.getUpperBound()-lIssueReal.getLowerBound())*(double)(optionInd)/(double)(lIssueReal.getNumberOfDiscretizationSteps()));
				standardBid.setValue(lIssueReal.getNumber(), pValue);
				double utility = getUtility(standardBid);
				getBids.put(standardBid, utility);
				break;
			case INTEGER:
				IssueInteger lIssueInteger = (IssueInteger)lIssue;
				int optionIndex2 = lIssueInteger.getLowerBound() + randomnr.nextInt(lIssueInteger.getUpperBound()-lIssueInteger.getLowerBound());
				Value pValue2 = new ValueInteger(optionIndex2);
				standardBid.setValue(lIssueInteger.getNumber(), pValue2);
				double utility2 = getUtility(standardBid);
				getBids.put(standardBid, utility2);
				break;
			default: throw new Exception("issue type "+lIssue.getType()+" not supported by AgentMR");
			}
		}

		return getBids;
	}

	/**
	 * This function determines the accept probability for an offer. At t=0 it
	 * will prefer high-utility offers. As t gets closer to 1, it will accept
	 * lower utility offers with increasing probability. it will never accept
	 * offers with utility 0.
	 *
	 * @param u
	 *            is the utility
	 * @param t
	 *            is the time as fraction of the total available time (t=0 at
	 *            start, and t=1 at end time)
	 * @return the probability of an accept at time t
	 * @throws Exception
	 *             if you use wrong values for u or t.
	 *
	 */
	double Paccept(double u, double t1) throws Exception {
		double t = t1 * t1 * t1; // steeper increase when deadline approaches.
		if (u < 0 || u > 1.05)
			throw new Exception("utility " + u + " outside [0,1]");
		// normalization may be slightly off, therefore we have a broad boundary
		// up to 1.05
		if (t < 0 || t > 1)
			throw new Exception("time " + t + " outside [0,1]");
		if (u > 1.)
			u = 1;
		if (t == 0.5)
			return u;
		return (u - 2. * u * t + 2. * (-1. + t + Math.sqrt(sq(-1. + t) + u
				* (-1. + 2 * t))))
				/ (-1. + 2 * t);
	}

	double sq(double x) {
		return x * x;
	}
}
