package Rumba;

import java.util.LinkedList;
import java.util.TreeMap;

import negotiator.Agent;
import negotiator.Bid;
import negotiator.actions.Accept;
import negotiator.actions.Action;
import negotiator.actions.Offer;
import negotiator.utility.UtilitySpace;

public class Logic {
	private static final double THRESHOLD_UTILITY = 0.01;
	private static final int MAX_OFFERS = 4;
	
	private final Agent agent;
	private final UtilitySpace utilitySpace;
	private final BidSpace bidSpace;
	private final ThresholdCalculator thresholdCalculator;
	private BidCollection ourBids;
	private BidCollection opponentBids;
	private LinkedList<EvaluatedBid> offerQueue;
	
	private EvaluatedBid opponentBestBid = null;
	
	private double minOfferUtility = 1.0;
	
	public Logic(Agent agent, 
			UtilitySpace utilitySpace, 
			BidSpace bidSpace, 
			ThresholdCalculator thresholdCalculator) {
		this.agent = agent;
		this.utilitySpace = utilitySpace;
		this.bidSpace = bidSpace;
		this.thresholdCalculator = thresholdCalculator;
		
		this.ourBids = new BidCollection();
		this.opponentBids = new BidCollection();
		this.offerQueue = new LinkedList<EvaluatedBid>();
		
		this.offerQueue.add(this.bidSpace.maxBid());
	}
	
	public void setOpponentOffer(Offer offer) {
		Bid offerBid = offer.getBid();
		
		double utility = 0;
		try {
			utility = this.utilitySpace.getUtility(offerBid);
		} catch (Exception e) {
		}
		
		EvaluatedBid opponentBid = new EvaluatedBid(offerBid, utility);
		this.opponentBids.addBid(opponentBid);
		
		try {
			if (this.opponentBestBid == null) {
				this.opponentBestBid = opponentBid;;
			}
			else if( opponentBid.getUtility() > this.opponentBestBid.getUtility()){
				this.opponentBestBid = opponentBid;
			}
			
			this.opponentBestBid = this.bidSpace.floorBid(opponentBestBid);	
		}
		catch(Exception ex){
		}
	}
	
	public Action chooseAction() {
		Action newAction = null;
		LinkedList<EvaluatedBid> offers = this.offerQueue;

		try
		{
			if (offers.isEmpty()) {
				addOffers(offers);
			}
			
			EvaluatedBid opponentLastBid = this.opponentBids.lastBid();

			// We should never get here
			if (offers.isEmpty()) {
				newAction = chooseActionOnError(opponentLastBid);
				return newAction;
			}
			
			boolean shouldAccept = 
				((opponentLastBid != null) && 
				(opponentLastBid.getUtility() > minOfferUtility ||
						offers.getFirst().getUtility() <= opponentLastBid.getUtility()));
						
			if (shouldAccept) {
				newAction = createAccept();
			}
			else {
				EvaluatedBid bid = offers.remove();
				newAction = createOffer(bid);
			}
		}
		catch (Exception e) {
			//System.out.println("Error: " + e + " stack: " + e.getStackTrace());
			newAction = createAccept();
		}
		
		return newAction;
	}

	private void addOffers(LinkedList<EvaluatedBid> offers) {
	
		double threshold = this.thresholdCalculator.getThreshold();
		TreeMap<Double, EvaluatedBid> pending = new TreeMap<Double,EvaluatedBid>();
		EvaluatedBid bidToTry = this.bidSpace.lowerBid(ourBids.lastBid());
		
		if (bidToTry.getUtility() >= threshold) {
			pending.put(bidToTry.getUtility(),bidToTry);
		}
		else {
			bidToTry = ourBids.randomBid();
			pending.put(bidToTry.getUtility(), bidToTry);
		}

		double initialUtility = bidToTry.getUtility();
		bidToTry = this.bidSpace.lowerBid(initialUtility);
		double nextUtility = bidToTry.getUtility();
		
		while( nextUtility >= threshold && (initialUtility-nextUtility) < THRESHOLD_UTILITY ) {
			pending.put(nextUtility,bidToTry);
			bidToTry = this.bidSpace.lowerBid(nextUtility);
			nextUtility = bidToTry.getUtility();
		}
		
		if (pending.size() <= MAX_OFFERS) {
			offers.addAll(pending.values());
		}
		else {
			int addedSofar = 0;
			EvaluatedBid bestBid = null;

			while (addedSofar <= MAX_OFFERS) {
				bestBid = pending.lastEntry().getValue();
				
				for(EvaluatedBid e : pending.values()) {
					if(e.getUtility() > bestBid.getUtility()){
						bestBid = e;
					}
				}
				offers.add(bestBid);
				pending.remove(bestBid.getUtility());
				addedSofar ++;
			}
		}
		
		if (offers.getFirst().getUtility() < this.opponentBestBid.getUtility()) {
			offers.addFirst(this.opponentBestBid);
		}
	}
	
	private Action chooseActionOnError(EvaluatedBid opponentLastBid) {
		EvaluatedBid randomBid = this.bidSpace.randomBid();
		
		if (randomBid == null) {
			return createAccept();
		}
		else if (opponentLastBid != null && randomBid.getUtility() <= opponentLastBid.getUtility()) {
			return createAccept();
		}
		else {
			return createOffer(randomBid);
		}
	}
	
	private Action createOffer(EvaluatedBid bid) {
		Offer offer = new Offer(this.agent.getAgentID(), bid.getBid());
		if (bid.getUtility() < minOfferUtility) {
			minOfferUtility = bid.getUtility(); 
		}
		ourBids.addBid(bid);
		return offer;
	}
	
	private Action createAccept() {
		return new Accept(this.agent.getAgentID());
	}
}
