package agents.AgentSmith;

import negotiator.Agent;
import negotiator.Bid;
import negotiator.BidIterator;
import negotiator.actions.Accept;
import negotiator.actions.Action;
import negotiator.actions.EndNegotiation;
import negotiator.actions.Offer;
import negotiator.issue.Issue;

/** 
 */
public class AgentSmith extends Agent 
{
	// max utility available
	private double maxUtility;
	
	private Action lastOpponentAction = null;
		
	public AgentSmith() 
	{
	}

	public static String getVersion() 
	{ 
		return "1.0"; 
	}

	public void init() 
	{
		try
		{
			lastOpponentAction = null;
			maxUtility = utilitySpace.getUtility(utilitySpace.getMaxUtilityBid());
		}
		catch (Exception e)
		{
			Double reservationValue = utilitySpace.getReservationValue();
			if (reservationValue == null)
			{
				maxUtility = utilitySpace.getReservationValue();
			}
			else
			{
				maxUtility = Double.MAX_VALUE;
			}
			e.printStackTrace();
		}
	}
	
	/**
	 * informs you which action the opponent did
	 * @param opponentAction
	 */
	public void ReceiveMessage(Action opponentAction) 
	{
		lastOpponentAction = opponentAction;
	}

	/**
     * this function is called after ReceiveMessage,
     * with an Offer-action.
     * @return (should return) the bid-action the agent wants to make.
     */
	public Action chooseAction() 
	{
		try 
		{
			if (lastOpponentAction == null) 
			{
				return getNextBid();
			} 
			else if (lastOpponentAction instanceof EndNegotiation) 
			{
				return null;
			} 
			else if (lastOpponentAction instanceof Offer) 
			{	
				Offer offer = (Offer) lastOpponentAction;
				Bid currentBid = offer.getBid();
				
				// we only willing to consider full bids
				if (!utilitySpace.constraintsViolated(currentBid) && isFullBid(currentBid))
				{
					// get the utility from the bid
					double ourUtility = utilitySpace.getUtility(currentBid);
					// the status quo is the outcome when the negotiation will timeout so it will have the full effect of time
					Double statusQuo = utilitySpace.getReservationValue();
					
					// if the utility that we will get make sense
					if (statusQuo == null || ourUtility >= statusQuo) 
					{		
						double minimalUtilityAcceptable = 
								Math.max(statusQuo == null ? maxUtility / 2.0 : statusQuo, 
									 maxUtility * (1.0 - timeline.getTime()*0.2));
						// if its acceptable so accept it
						if (utilitySpace.getUtility(currentBid) >= minimalUtilityAcceptable)
						{
							Accept answer = new Accept(getAgentID());
							return answer;
						}
						return getNextBid();
					} 
					return getNextBid();
				}
				return getNextBid();
			}
			else if (lastOpponentAction instanceof Accept) 
			{
				// nothing really to do
				return null;
			} 
			else 
			{
				// unknown action - get next bid
				return getNextBid();
			}
		}
		catch ( Exception e) 
		{
			System.err.println("AsyncAgent: Exception in ChooseAction:"+e.getMessage());
			e.printStackTrace();
			return null;
		}
	}
	
	// gets the bids in descending order of our utility and our preferences
	// and that are smaller that our last utility with delta
	private Offer getNextBid() throws Exception
	{
		BidIterator bidIter = new BidIterator(utilitySpace.getDomain());
		double minUtility = Double.POSITIVE_INFINITY;
		Bid minBid = null;
		Double statusQuo = utilitySpace.getReservationValue();
		double minimalUtilityAcceptable = 
			Math.max(statusQuo == null ? maxUtility / 2.0 : statusQuo, 
				 maxUtility * (1.0 - timeline.getTime()*0.2));
		while (bidIter.hasNext()) 
		{
			Bid bid = bidIter.next();
			// we offer only full bids, partial can't really help
			// find the min bid with utility above the minimum acceptable
			if (isFullBid(bid))
			{
				double currentBidUtility = utilitySpace.getUtility(bid);
				if (currentBidUtility < minUtility && currentBidUtility >= minimalUtilityAcceptable)
				{
					minUtility = currentBidUtility;
					minBid = bid;
				}
			}
		}
		if (minBid == null)
		{
			minBid = utilitySpace.getDomain().getRandomBid();
		}
		
		// now we offer it
		Offer ourOffer = new Offer(getAgentID(),minBid);
		return ourOffer;
	}
	
	public String getName()
	{
		return "AgentSmith";
	}
	
	private boolean isFullBid(Bid bid)
	{
		for (Issue issue : utilitySpace.getDomain().getIssues())
		{
			try
			{
				bid.getValue(issue.getNumber());
			}
			catch (Exception e) 
			{
				return false;
			}
		}
		return true;
	}
}

