package agents.TheNegotiator;

import negotiator.actions.Accept;
import negotiator.actions.Action;
import negotiator.utility.UtilitySpace;

/**
 * The Acceptor class is used to decide when to accept a bid.
 * 
 * @author Alex Dirkzwager, Mark Hendrikx, Julian de Ruiter
 */
public class Acceptor {
	
	// utilityspace of the negotiation
	private UtilitySpace utilitySpace;
	// reference to the bidscollection
	private BidsCollection bidsCollection;
	
	/**
	 * Creates an Acceptor-object which determines which offers should be accepted
	 * during the negotiation.
	 * 
	 * @param utilitySpace
	 * @param bidsCollection of all possible bids (for us) and the partner bids
	 */
	public Acceptor(UtilitySpace utilitySpace, BidsCollection bidsCollection) {
		this.utilitySpace = utilitySpace;
		this.bidsCollection = bidsCollection;
	}

	/**
	 * Determine if it is wise accept an offer based on the X last partner bids.
	 * 
	 * @param lastX determines the amount of previous partner offers taken into account
	 * @param utility of the current partner bid
	 * @return if it is wise to accept the offer
	 */
	private boolean acceptBasedOnMV(int lastX, double utility) {
		// if more bids are asked than are available
		if (lastX >= bidsCollection.getPartnerBids().size()) {
			lastX = bidsCollection.getPartnerBids().size() - 1;
		}
		
		// calculate statistics of the opponent bids
		double mean = bidsCollection.meanPartnerBids(lastX);
		double variance = bidsCollection.variancePartnerBids(lastX);
		double standardDeviation = Math.sqrt(variance);
		
		// the chance to get a bid with an utility above mean + 1.25 * standard 
		// deviation is about 10% (assuming normal distribution of opponent bid utility,
		// which is more or less correct for most agents)
		return utility > (mean + 1.25 * standardDeviation);
	}
	
	/**
	 * Determine if it is wise to accept for a given phase on a given time.
	 * 
	 * @param phase of the negotiation
	 * @param time in negotiation
	 * @return if it is wise to accept
	 */
	public Action determineAccept(int phase, double threshold, double time, boolean discount) {
		Action action = null;
		double utility = 0;

		try {
			// effectively this will ensure that the utility is 0 if our agent is first
			if (bidsCollection.getPartnerBids().size() > 0) {
				utility = utilitySpace.getUtility(bidsCollection.getPartnerBid(0));
			}
		} catch (Exception e) {
			ErrorLogger.log("ACCEPTOR: error during calculation of utility.");
			ErrorLogger.log("Error on PartnerBid: " + bidsCollection.getPartnerBid(0));
		}
		
		switch (phase) {
			case 1:
				// only accept if the opponent bid is very good
				if (utility >= threshold) {
					action = new Accept();
				}
				break;
			case 2:
				// only accept if the bid is very good, and above the threshold
				if (acceptBasedOnMV(8, utility) && utility >= threshold) {
					action = new Accept();
				}
				break;
			case 3:
//					 only accept if the opponent bid is pretty good
				if (utility >= threshold) {
					action = new Accept();
				}
				break;
			case 4:
				// only accept if the time in phase 4 is nearly over and the opponent
				// bid is above the threshold. If all opponent bids are too low in the
				// time span, then it can happen that no agreement is reached...
				if (time > 0.985 && utility >= threshold) {
					action = new Accept();
				}
				break;
			default: 
				ErrorLogger.log("Unknown phase");
				break;
		}
		return action;
	}
}