package agents.TheNegotiator;

import java.util.ArrayList;

import negotiator.Timeline;

/**
 * The TimeManager class is used for time-related functions.
 * 
 * @author Alex Dirkzwager, Mark Hendrikx, Julian de Ruiter
 */
public class TimeManager {

	// used to store the timeline of the negotiation
	private Timeline timeline;
	// total time steps considered
	private final double totalTime = 180;
	// times when phases end for a non discount domain
	private final double[] endPhases = { 90.0 / totalTime, 150.0 / totalTime, 165.0 / totalTime, 1.0};
	// if it is a discount domain
	private boolean discount = false;
	// collection of bids
	private BidsCollection bidsCollection;
	// collection of threshold intervals which correspond to the phases
	private ArrayList<Pair<Double,Double>> thresArray = new ArrayList<Pair<Double, Double>>();
	// array for storing how many bids are in a certain threshold interval
	private final int propArray[] = {0, 0, 0, 0};
	// maximum bid utility within a threshold interval
	private final double maxThresArray[] = {0, 0, 0, 0, 0};
	
	/**
	 * Creates a TimeManager-object which stores the timeline of the negotiation.
	 *
	 * @param timeline of the negotiation
	 */
	public TimeManager(Timeline timeline, boolean discount, BidsCollection bidsCollection) {
		this.timeline = timeline;
		this.discount = discount;
		this.bidsCollection = bidsCollection;

		// creating the threshold intervals...
		thresArray.add(new Pair<Double, Double>(0.825, 1.1)); // 1.1 because of GENIUS rounding errors :)
		thresArray.add(new Pair<Double, Double>(0.65 ,0.825));
		thresArray.add(new Pair<Double, Double>(0.475 , 0.65));
		thresArray.add(new Pair<Double, Double>(0.3, 0.475));

		// switch to the strategy for strong discounts
		if (this.discount) {
			calculatePropArray();
			calculateEndPhases();
		}
	}

	/**
	 * Returns the current phase of the negotiation.
	 * 
	 * @return phase of the negotiation
	 */
	public int getPhase() {
		int phase = 1;
		double time = timeline.getTime();
		
		if (time > endPhases[2]) {
			phase = 4;
		} else if (time > endPhases[1]) {
			phase = 3;
		} else if (time > endPhases[0]) {
			phase = 2;
		}
		return phase;
	}
	
	/**
	 * Calculate how many possible bids are within a certain threshold interval.
	 * This is done for all the bins (phases).
	 */
	public void calculatePropArray() {
		ArrayList<UTBid> posBids = bidsCollection.getPossibleBids();
		
		for (int i = 0; i < posBids.size(); i++) {
			double util = posBids.get(i).getUtility();

			// calculate if a utility of a bid is within a certain interval
			if (util >= thresArray.get(0).getFirst() && util < thresArray.get(0).getSecond()) {
				propArray[0]++;
			} else if (util >= thresArray.get(1).getFirst() && util < thresArray.get(1).getSecond()) {
				propArray[1]++;
			} else if (util >= thresArray.get(2).getFirst() && util < thresArray.get(2).getSecond()) {
				propArray[2]++;
			} else if (util >= thresArray.get(3).getFirst() && util < thresArray.get(3).getSecond()) {
				propArray[3]++;
			}
		}
		
		// find the maximum possible utility within a bin (plus the lowest bid of the last bin)
		ArrayList<UTBid> bidsCol = bidsCollection.getPossibleBids();
		maxThresArray[0] = bidsCol.get(0).getUtility();
		maxThresArray[1] = bidsCol.get(propArray[0]).getUtility();
		maxThresArray[2] = bidsCol.get(propArray[1]).getUtility();
		maxThresArray[3] = bidsCol.get(propArray[2]).getUtility();
		maxThresArray[4] = bidsCol.get(propArray[3] - 1).getUtility();
		
	}
	
	/**
	 * Calculates the time which should be spend on each phase based on
	 * the distribution of the utilities of the bids.
	 */
	public void calculateEndPhases() {
		for (int i = 0; i < propArray.length; i++) {
			double percentage = (double)propArray[i] / (double)bidsCollection.getPossibleBids().size();
			if (i > 0) {
				percentage += endPhases[i - 1];
			}
			endPhases[i] = percentage;
		}
	}
	
	/**
	 * Returns the time dependent threshold which specifies how good a 
	 * bid of an opponent should be to be accepted. This threshold is 
	 * also used as a minimum for the utility of the bid of our agent.
	 * 
	 * @return threshold
	 */
	public double getThreshold() {
		double time = timeline.getTime();
		int phase = getPhase();
		double threshold;
		
		// for a domain with a strong discount
		if (discount) {
			switch (phase) {
				// do a linear interpolation of the thresholdvalues between phases
				case 1: threshold = maxThresArray[0] - ((time / endPhases[0]) * (maxThresArray[0] - maxThresArray[1])); break;
				case 2: threshold = maxThresArray[1] - (((time - endPhases[0]) / (endPhases[1])) * (maxThresArray[1] - maxThresArray[2])); break;
				case 3: threshold = maxThresArray[2] - (((time - endPhases[1]) / (endPhases[2])) * (maxThresArray[2] - maxThresArray[3])); break;
				case 4: threshold = maxThresArray[3] - (((time - endPhases[2]) / (endPhases[3])) * (maxThresArray[3] - maxThresArray[4])); break;
				default: ErrorLogger.log("Unknown phase: " + phase); threshold = 1; break;
			}
		} else { // for normal domains
			switch (phase) {
        		case 1: threshold = 1 - ((time / endPhases[0]) * 0.10); break;
        		case 2: threshold = 0.90 - (((time - endPhases[0]) / (endPhases[1] - endPhases[0])) * 0.10); break;
        		case 3: threshold = 0.80 - (((time - endPhases[1]) / (endPhases[2] - endPhases[1])) * 0.10); break;
        		case 4: threshold = 0.70 - (((time - endPhases[2]) / (endPhases[3] - endPhases[2])) * 0.20); break;
        		default: ErrorLogger.log("Unknown phase: " + phase); threshold = 1; break;
			}
		}
		return threshold;
	}
}