package travelingsalesman;

import negotiator.analysis.BidPoint;
import negotiator.utility.UtilitySpace;

/**
 * Class used to estimate the Kalai point. This approach works well
 * with Bayesian models in which all bids are used in the updating process.
 * 
 * @author Mark Hendrikx
 */
public class DomainAnalyzer {
	
	// when an opponent model is given, the model is used in the Kalai estimation
	private OpponentModel opponentModel;
	// the utility space of our agent
	private UtilitySpace ownUtilSpace;
	// a reference to the opponent history
	private BidHistory opponentHistory;
	// size of the domain
	private long domainSize;
	// estimate of the amount of bids which should be done before the opponent model is reliable
	private static int MINIMUM_BIDS = 200;
	// default value for Kalai which is used when the opponent model is unavailable or unreliable
	private static double DEFAULT_KALAI = 0.7;
	
	/**
	 * Set the domain analyzer variables and determine the domain size.
	 * 
	 * @param ownUtilSpace utility space of our agen
	 * @param opponentModel used to estimate the opponent's preference profile
	 * @param opponentHistory
	 */
	public DomainAnalyzer(UtilitySpace ownUtilSpace, OpponentModel opponentModel, BidHistory opponentHistory) {
		this.opponentModel = opponentModel;
		this.ownUtilSpace = ownUtilSpace;
		this.opponentHistory = opponentHistory;
		domainSize = ownUtilSpace.getDomain().getNumberOfPossibleBids();
	}
	
	/**
	 * Calculates the Kalai point by optionally using an opponent model.
	 * When an opponent model in unavailable, unreliable, or the domain too large,
	 * a default value is used.
	 * 
	 * @return coordinate of the Kalai point from our utilityspace
	 */
	public double calculateKalaiPoint() {
		double kalaiPoint = DEFAULT_KALAI;
		if (opponentModel != null && opponentModel.isCompleteModel() && 
				opponentHistory.size() > MINIMUM_BIDS && domainSize < 200000) {
			BidSpace space;
			try {
				
				space = new BidSpace(ownUtilSpace, opponentModel.getOpponentUtilitySpace(), true);
				BidPoint kalai = space.getKalaiSmorodinsky();
				kalaiPoint = kalai.utilityA;
			} catch (Exception e) {
				e.printStackTrace();
			}
		}
		return kalaiPoint;
	}
}